@@ -33,6 +33,7 @@
 #include "common/utils.h"
 #include "kernel-shared/volumes.h"
 #include "common/open-utils.h"
+#include "common/parse-utils.h"
 #include "common/device-scan.h"
 #include "common/help.h"
 #include "common/box.h"
@@ -58,6 +59,10 @@ static int update_seeding_flag(struct btrfs_root *root, int set_flag)
 						device);
 			return 1;
 		}
+		if (btrfs_super_log_root(disk_super)) {
+			error("filesystem with dirty log detected, not setting seed flag");
+			return 1;
+		}
 		super_flags |= BTRFS_SUPER_FLAG_SEEDING;
 	} else {
 		if (!(super_flags & BTRFS_SUPER_FLAG_SEEDING)) {
@@ -186,6 +191,314 @@ static int set_metadata_uuid(struct btrfs_root *root, const char *uuid_string)
 	return btrfs_commit_transaction(trans, root);
 }
 
+static int delete_csum_items(struct btrfs_trans_handle *trans,
+			     struct btrfs_fs_info *fs_info)
+{
+	struct btrfs_root *root = btrfs_csum_root(fs_info, 0);
+	struct btrfs_path path;
+	struct btrfs_key key;
+	int nr;
+	int ret;
+
+	btrfs_init_path(&path);
+
+	key.objectid = BTRFS_EXTENT_CSUM_OBJECTID;
+	key.type = BTRFS_EXTENT_CSUM_KEY;
+	key.offset = 0;
+
+	while (1) {
+		ret = btrfs_search_slot(trans, root, &key, &path, -1, 1);
+		if (ret < 0)
+			goto out;
+
+		nr = btrfs_header_nritems(path.nodes[0]);
+		if (!nr)
+			break;
+
+		path.slots[0] = 0;
+		ret = btrfs_del_items(trans, root, &path, 0, nr);
+		if (ret)
+			goto out;
+
+		btrfs_release_path(&path);
+	}
+
+	ret = 0;
+out:
+	btrfs_release_path(&path);
+	return ret;
+}
+
+static int change_extents_csum(struct btrfs_fs_info *fs_info, int csum_type)
+{
+	struct btrfs_root *root = btrfs_extent_root(fs_info, 0);
+	struct btrfs_path path;
+	struct btrfs_key key = {0, 0, 0};
+	int ret = 0;
+
+	btrfs_init_path(&path);
+	/*
+	 * Here we don't use transaction as it will takes a lot of reserve
+	 * space, and that will make a near-full btrfs unable to change csums
+	 */
+	ret = btrfs_search_slot(NULL, root, &key, &path, 0, 0);
+	if (ret < 0)
+		goto out;
+
+	while (1) {
+		struct btrfs_extent_item *ei;
+		struct extent_buffer *eb;
+		u64 flags;
+		u64 bytenr;
+
+		btrfs_item_key_to_cpu(path.nodes[0], &key, path.slots[0]);
+		if (key.type != BTRFS_EXTENT_ITEM_KEY &&
+		    key.type != BTRFS_METADATA_ITEM_KEY)
+			goto next;
+		ei = btrfs_item_ptr(path.nodes[0], path.slots[0],
+				    struct btrfs_extent_item);
+		flags = btrfs_extent_flags(path.nodes[0], ei);
+		if (!(flags & BTRFS_EXTENT_FLAG_TREE_BLOCK))
+			goto next;
+
+		bytenr = key.objectid;
+		eb = read_tree_block(fs_info, bytenr, 0);
+		if (IS_ERR(eb)) {
+			error("failed to read tree block: %llu", bytenr);
+			ret = PTR_ERR(eb);
+			goto out;
+		}
+		/* Only rewrite block */
+		/* printf("CSUM: start %llu\n", eb->start); */
+		ret = write_tree_block(NULL, fs_info, eb);
+		free_extent_buffer(eb);
+		if (ret < 0) {
+			error("failed to change csum of tree block: %llu", bytenr);
+			goto out;
+		}
+next:
+		ret = btrfs_next_item(root, &path);
+		if (ret < 0)
+			goto out;
+		if (ret > 0) {
+			ret = 0;
+			goto out;
+		}
+	}
+
+out:
+	btrfs_release_path(&path);
+	return ret;
+}
+
+static int change_devices_csum(struct btrfs_root *root, int csum_type)
+{
+	struct btrfs_fs_info *fs_info = root->fs_info;
+	struct btrfs_path path;
+	struct btrfs_key key = {0, 0, 0};
+	int ret = 0;
+
+	btrfs_init_path(&path);
+	/* No transaction again */
+	ret = btrfs_search_slot(NULL, root, &key, &path, 0, 0);
+	if (ret < 0)
+		goto out;
+
+	while (1) {
+		btrfs_item_key_to_cpu(path.nodes[0], &key, path.slots[0]);
+		if (key.type != BTRFS_DEV_ITEM_KEY ||
+		    key.objectid != BTRFS_DEV_ITEMS_OBJECTID)
+			goto next;
+		/* Only rewrite block */
+		ret = write_tree_block(NULL, fs_info, path.nodes[0]);
+		if (ret < 0)
+			goto out;
+next:
+		ret = btrfs_next_item(root, &path);
+		if (ret < 0)
+			goto out;
+		if (ret > 0) {
+			ret = 0;
+			goto out;
+		}
+	}
+out:
+	btrfs_release_path(&path);
+	return ret;
+}
+
+static int populate_csum(struct btrfs_trans_handle *trans,
+			 struct btrfs_fs_info *fs_info, char *buf, u64 start,
+			 u64 len)
+{
+	u64 offset = 0;
+	u64 sectorsize;
+	int ret = 0;
+
+	while (offset < len) {
+		sectorsize = fs_info->sectorsize;
+		ret = read_data_from_disk(fs_info, buf, start + offset,
+					  &sectorsize, 0);
+		if (ret)
+			break;
+		ret = btrfs_csum_file_block(trans, start + len, start + offset,
+				buf, sectorsize);
+		if (ret)
+			break;
+		offset += sectorsize;
+	}
+	return ret;
+}
+
+static int fill_csum_tree_from_extent(struct btrfs_fs_info *fs_info)
+{
+	struct btrfs_root *extent_root = btrfs_extent_root(fs_info, 0);
+	struct btrfs_trans_handle *trans;
+	struct btrfs_path path;
+	struct btrfs_extent_item *ei;
+	struct extent_buffer *leaf;
+	char *buf;
+	struct btrfs_key key;
+	int ret;
+
+	trans = btrfs_start_transaction(extent_root, 1);
+	if (trans == NULL) {
+		/* fixme */
+		printf("cannot start transaction\n");
+		return -EINVAL;
+	}
+
+	btrfs_init_path(&path);
+	key.objectid = 0;
+	key.type = BTRFS_EXTENT_ITEM_KEY;
+	key.offset = 0;
+	ret = btrfs_search_slot(NULL, extent_root, &key, &path, 0, 0);
+	if (ret < 0) {
+		btrfs_release_path(&path);
+		return ret;
+	}
+
+	buf = malloc(fs_info->sectorsize);
+	if (!buf) {
+		btrfs_release_path(&path);
+		return -ENOMEM;
+	}
+
+	ret = delete_csum_items(trans, fs_info);
+	if (ret) {
+		error("unable to delete all checksum items: %d", ret);
+		return -EIO;
+	}
+
+	while (1) {
+		if (path.slots[0] >= btrfs_header_nritems(path.nodes[0])) {
+			ret = btrfs_next_leaf(extent_root, &path);
+			if (ret < 0)
+				break;
+			if (ret) {
+				ret = 0;
+				break;
+			}
+		}
+		leaf = path.nodes[0];
+
+		btrfs_item_key_to_cpu(leaf, &key, path.slots[0]);
+		if (key.type != BTRFS_EXTENT_ITEM_KEY) {
+			path.slots[0]++;
+			continue;
+		}
+
+		ei = btrfs_item_ptr(leaf, path.slots[0], struct btrfs_extent_item);
+		if (!(btrfs_extent_flags(leaf, ei) & BTRFS_EXTENT_FLAG_DATA)) {
+			path.slots[0]++;
+			continue;
+		}
+
+		ret = populate_csum(trans, fs_info, buf, key.objectid, key.offset);
+		if (ret)
+			break;
+		path.slots[0]++;
+	}
+
+	btrfs_release_path(&path);
+	free(buf);
+
+	/* dont' commit if thre's error */
+	ret = btrfs_commit_transaction(trans, extent_root);
+
+	return ret;
+}
+
+static int rewrite_checksums(struct btrfs_root *root, int csum_type)
+{
+	struct btrfs_fs_info *fs_info = root->fs_info;
+	struct btrfs_super_block *disk_super;
+	struct btrfs_trans_handle *trans;
+	u64 super_flags;
+	int ret;
+
+	disk_super = root->fs_info->super_copy;
+	super_flags = btrfs_super_flags(disk_super);
+
+	/* FIXME: Sanity checks */
+	if (0) {
+		fprintf(stderr,
+			"UUID rewrite in progress, cannot change fsid\n");
+		return 1;
+	}
+
+	fs_info->force_csum_type = csum_type;
+
+	/* Step 1 sets the in progress flag, no other change to the sb */
+	printf("Set superblock flag CHANGING_CSUM\n");
+	trans = btrfs_start_transaction(root, 1);
+	super_flags |= BTRFS_SUPER_FLAG_CHANGING_CSUM;
+	btrfs_set_super_flags(disk_super, super_flags);
+	ret = btrfs_commit_transaction(trans, root);
+	if (ret < 0)
+		return ret;
+
+	/* Change extents first */
+	printf("Change fsid in extents\n");
+	ret = change_extents_csum(fs_info, csum_type);
+	if (ret < 0) {
+		error("failed to change csum of metadata: %d", ret);
+		goto out;
+	}
+
+	/* Then devices */
+	printf("Change csum in chunk tree\n");
+	ret = change_devices_csum(fs_info->chunk_root, csum_type);
+	if (ret < 0) {
+		error("failed to change UUID of devices: %d", ret);
+		goto out;
+	}
+
+	/* DATA */
+	printf("Change csum of data blocks\n");
+	ret = fill_csum_tree_from_extent(fs_info);
+	if (ret < 0)
+		goto out;
+
+	/* Last, change fsid in super */
+	ret = write_all_supers(fs_info);
+	if (ret < 0)
+		goto out;
+
+	/* All checksums done, drop the flag, super block csum will get updated */
+	printf("Clear superblock flag CHANGING_CSUM\n");
+	super_flags = btrfs_super_flags(fs_info->super_copy);
+	super_flags &= ~BTRFS_SUPER_FLAG_CHANGING_CSUM;
+	btrfs_set_super_flags(fs_info->super_copy, super_flags);
+	btrfs_set_super_csum_type(disk_super, csum_type);
+	ret = write_all_supers(fs_info);
+	printf("Checksum change finished\n");
+out:
+	/* check errors */
+
+	return ret;
+}
+
 static int set_super_incompat_flags(struct btrfs_root *root, u64 flags)
 {
 	struct btrfs_trans_handle *trans;
@@ -481,6 +794,11 @@ static void print_usage(void)
 	printf("  general:\n");
 	printf("\t-f          allow dangerous operations, make sure that you are aware of the dangers\n");
 	printf("\t--help      print this help\n");
+#ifdef EXPERIMENTAL
+	printf("\nEXPERIMENTAL FEATURES:\n");
+	printf("  checksum changes:\n");
+	printf("\t--csum CSUM	switch checksum for data and metadata to CSUM\n");
+#endif
 }
 
 int BOX_MAIN(btrfstune)(int argc, char *argv[])
@@ -493,14 +811,19 @@ int BOX_MAIN(btrfstune)(int argc, char *argv[])
 	u64 seeding_value = 0;
 	int random_fsid = 0;
 	int change_metadata_uuid = 0;
+	int csum_type = -1;
 	char *new_fsid_str = NULL;
 	int ret;
 	u64 super_flags = 0;
 	int fd = -1;
 
 	while(1) {
+		enum { GETOPT_VAL_CSUM = 256 };
 		static const struct option long_options[] = {
 			{ "help", no_argument, NULL, GETOPT_VAL_HELP},
+#ifdef EXPERIMENTAL
+			{ "csum", required_argument, NULL, GETOPT_VAL_CSUM },
+#endif
 			{ NULL, 0, NULL, 0 }
 		};
 		int c = getopt_long(argc, argv, "S:rxfuU:nmM:", long_options, NULL);
@@ -541,6 +864,15 @@ int BOX_MAIN(btrfstune)(int argc, char *argv[])
 			ctree_flags |= OPEN_CTREE_IGNORE_FSID_MISMATCH;
 			change_metadata_uuid = 1;
 			break;
+#ifdef EXPERIMENTAL
+		case GETOPT_VAL_CSUM:
+			ctree_flags |= OPEN_CTREE_SKIP_CSUM_CHECK;
+			csum_type = parse_csum_type(optarg);
+			warning("Switching checksums is experimental, do not use for valuable data!");
+			printf("Switch csum to %s\n",
+					btrfs_super_csum_name(csum_type));
+			break;
+#endif
 		case GETOPT_VAL_HELP:
 		default:
 			print_usage();
@@ -558,7 +890,7 @@ int BOX_MAIN(btrfstune)(int argc, char *argv[])
 		return 1;
 	}
 	if (!super_flags && !seeding_flag && !(random_fsid || new_fsid_str) &&
-	    !change_metadata_uuid) {
+	    !change_metadata_uuid && csum_type == -1) {
 		error("at least one option should be specified");
 		print_usage();
 		return 1;
@@ -635,6 +967,12 @@ int BOX_MAIN(btrfstune)(int argc, char *argv[])
 		total++;
 	}
 
+	if (csum_type != -1) {
+		/* TODO: check conflicting flags */
+		printf("Proceed to switch checksums\n");
+		ret = rewrite_checksums(root, csum_type);
+	}
+
 	if (change_metadata_uuid) {
 		if (seeding_flag) {
 			fprintf(stderr,
